@@ -44,5 +44,5 @@ int jpegIdent();
 int imIdent();
 #ifdef USE_PNG
 int pngIdent();
-int pngLoad();
+Image *pngLoad();
 #endif
