@@ -26,10 +26,10 @@
  * SUCH DAMAGE.
  */
 /*
- * $Id: mgp.c,v 1.136 2004/09/14 14:48:47 nishida Exp $
+ * $Id: mgp.c,v 1.145 2007/01/21 17:56:01 nishida Exp $
  */
 
-static char *mgp_version = "1.11a (20040915)";
+static char *mgp_version = "1.12a (20070123)";
 
 #include "mgp.h"
 #ifdef USE_SETLOCALE
@@ -497,6 +497,11 @@ genhtml(start_page)
 				{ NULL, NULL, NULL }};
 	int inum = 0;
 
+#ifdef USE_M17N
+	/* initaiize M17N library */
+	M17N_init();
+#endif
+
 	/* check image type */
 	if (htmlimage) {
 		for (inum = 0; *convdb[inum] != NULL; inum++) {
@@ -628,7 +633,8 @@ genhtml(start_page)
 					goto txtdone;
 				break;
 			case CTL_TEXT:
-				fprintf(txt, "%s", state.cp->ctc_value);
+				if (state.cp->ctc_value)
+					fprintf(txt, "%s", state.cp->ctc_value);
 				break;
 			case CTL_LINEEND:
 				fprintf(txt, "\n");
@@ -728,7 +734,7 @@ mgp_usage(name)
 	fprintf(stderr, "\t-Q <quality>: Set background image quality(0-100)\n");
 	fprintf(stderr, "\t-R: Do not perform automatic reload\n");
 	fprintf(stderr, "\t-S: Do not process directives that forks process (default)\n");
-	fprintf(stderr, "\t-U: Do process directives that forks process (unsecure mode)\n");
+	fprintf(stderr, "\t-U: Do process directives that forks process\n\t    or allow to use non-ASCII filenames (unsecure mode)\n");
 	fprintf(stderr, "\t-T <timestampfile>: Update timestampfile on page refresh\n");
 	fprintf(stderr, "\t-P: print stderr from image conversion tools (by default it's discarded)\n");
 	fprintf(stderr, "\t-V: Be verbose\n");
@@ -939,17 +945,21 @@ main_loop(start_page)
 				struct render_state tstate;
 				tstate = state;
 
-				if (number == 0)
-					number = 1;
-				if (state.cp && state.cp->ct_op == CTL_PAUSE){
+				if (number == 0 && state.cp && state.cp->ct_op == CTL_PAUSE) {
 					state_next(&tstate);
-				} else if (state.page + number
+				} else {
+				    if (number == 0)
+				        number = 1;
+
+				    if (state.page + number
 						<= maxpage) {
 						state_goto(&tstate, state.page + number, 0);
-				} else {
-					beep();
+				    } else {
+				        beep();
 					break;
+				    }
 				}
+			       
 
 				if (memcmp(&state, &tstate,
 						sizeof(state)) != 0) {
@@ -1129,10 +1139,6 @@ repaint:;
 			}
 			break;
 
-#if 1  /* by h.kakugawa@computer.org */
-		case MapNotify:
-		  goto repaint;
-#endif
 		case Expose:
 			if (e.xexpose.window != window) 
 				break;
@@ -1142,7 +1148,7 @@ repaint:;
 			/* compress expose event */
 			while (XEventsQueued(display, QueuedAfterReading) > 0) {
 				XPeekEvent(display, &ahead);
-				if (ahead.type != Expose)
+				if (ahead.type != Expose && ahead.type != ConfigureNotify)
 					break;
 				if (ahead.xexpose.window != window)
 					break;
@@ -1161,6 +1167,15 @@ repaint:;
 				((mgp_flag & FL_OVER) || 
 				e.xconfigure.window != RootWindow(display, screen)))
 				break;
+			/* compress expose event */
+			while (XEventsQueued(display, QueuedAfterReading) > 0) {
+				XPeekEvent(display, &ahead);
+				if (ahead.type != Expose && ahead.type != ConfigureNotify)
+					break;
+				if (ahead.xconfigure.window != window)
+					break;
+				XNextEvent(display, &e);	
+			}
 			if (window_width != e.xconfigure.width
 			 || window_height != e.xconfigure.height) {
 				struct ctrl *lastcp;
