@@ -26,7 +26,7 @@
  * SUCH DAMAGE.
  */
 /*
- * $Id: mgp.h,v 1.126 2001/02/03 18:46:50 nishida Exp $
+ * $Id: mgp.h,v 1.143 2007/01/07 09:41:53 nishida Exp $
  */
 
 #include <stdio.h>
@@ -61,6 +61,15 @@
 #include <X11/keysym.h>
 #include <X11/cursorfont.h>
 #include <X11/Xatom.h>
+#ifdef USE_XFT2
+#include <ft2build.h>
+#include FT_FREETYPE_H
+#include <fontconfig/fontconfig.h>
+#include <X11/Xft/Xft.h>
+#ifdef HAVE_ICONV
+#include <iconv.h>
+#endif
+#endif
 #include "image/xloadimage.h"
 
 #ifndef __P
@@ -95,20 +104,30 @@
 #undef	TTY_KEYINPUT
 #endif
 
+#ifdef USE_M17N
+#include <m17n-gui.h>
+#include <m17n-misc.h>
+#include <m17n-X.h>
+#endif
+
 #define DEFAULT_FORE	"yellow"
 #define DEFAULT_BACK	"black"
 #define PAGELIST_FONT	"a14"
 #define PAGELIST_KFONT	"k14"
 
 #define DEFAULT_CHARSIZE	10	/* 10% of height */
+#define DEFAULT_SUPSCALE	0.6
+#define DEFAULT_SUPOFF		0.4
+#define DEFAULT_SUBOFF		0.15
 #define DEFAULT_HGAP		0
 #define DEFAULT_VGAP		15
 #define DEFAULT_BQUALITY	100
+#define DEFAULT_OPAQUE		100	
 
 #define XLFD_HYPHEN	14
 #define FONT_FORMAT	"-*-%s-*-*-%s-*-*-*-*-*-%s"
 #define DEFAULT_X_FONT	"times-medium-r"
-#define CUTIN_DELAY	15000
+#define CUTIN_DELAY	5000
 #define SHRINK_DELAY	00500
 #define EXEC_DELAY  20000
 
@@ -168,6 +187,8 @@
 #define	FL_NOSTDIN	0x2000
 #define	FL_GLYPHEDGE	0x4000
 #define	FL_FRDCACHE		0x8000
+#define	FL_NOXFT		0x10000
+#define	FL_NOM17N		0x20000
 
 /* 	page attribute flags 	*/
 #define PGFLAG_NODEF	0x01	/* nodefault */
@@ -188,6 +209,17 @@ struct ctrl_int {
 	u_int ct_value;
 };
 
+struct ctrl_int2 {
+       u_int ct_value1;
+        u_int ct_value2;
+};
+
+struct ctrl_int3 {
+       u_int ct_value1;
+        u_int ct_value2;
+        u_int ct_value3;
+	};
+
 struct ctrl_long {
 	u_long ct_value;
 };
@@ -213,6 +245,8 @@ struct ctrl_image {
 	u_int ct_ximagesize;
 	u_int ct_yimagesize;
 	u_int ct_zoomflag;
+	u_int ct_raise;
+	u_int ct_rotate; /* +/-180, +/-90, 0, 270 */
 #define Z_XMASK		0x0f
 #define Z_YMASK		0xf0
 #define Z_YSHIFT	4
@@ -257,6 +291,13 @@ struct ctrl_icon {
 	u_int ct_size;   
 };  
 
+struct ctrl_area {
+	u_int ct_xoff;
+	u_int ct_width;
+	u_int ct_yoff;
+	u_int ct_height;
+};
+
 struct ctrl_pcache {
 	u_int ct_cflag;
 	u_int ct_cmode;
@@ -272,6 +313,8 @@ struct ctrl {
 	union {
 		struct ctrl_double ctrl_double;
 		struct ctrl_int ctrl_int;
+                struct ctrl_int2 ctrl_int2;
+                struct ctrl_int3 ctrl_int3;        
 		struct ctrl_long ctrl_long;
 		struct ctrl_char ctrl_char;
 		struct ctrl_char2 ctrl_char2;
@@ -280,12 +323,18 @@ struct ctrl {
 		struct ctrl_bar ctrl_bar;
 		struct ctrl_args ctrl_args;
 		struct ctrl_icon ctrl_icon;
+		struct ctrl_area ctrl_area;
 		struct ctrl_pcache ctrl_pcache;
 	} ct_val;
 };
 
 #define ctf_value	ct_val.ctrl_double.ct_value
 #define cti_value	ct_val.ctrl_int.ct_value
+#define cti2_value1	ct_val.ctrl_int2.ct_value1
+#define cti2_value2	ct_val.ctrl_int2.ct_value2
+#define cti3_value1	ct_val.ctrl_int3.ct_value1
+#define cti3_value2	ct_val.ctrl_int3.ct_value2
+#define cti3_value3	ct_val.ctrl_int3.ct_value3
 #define ctl_value	ct_val.ctrl_long.ct_value
 #define ctc_value	ct_val.ctrl_char.ct_value
 #define ctc2_value1	ct_val.ctrl_char2.ct_value1
@@ -295,6 +344,8 @@ struct ctrl {
 #define ctm_ximagesize	ct_val.ctrl_image.ct_ximagesize
 #define ctm_yimagesize	ct_val.ctrl_image.ct_yimagesize
 #define ctm_zoomflag	ct_val.ctrl_image.ct_zoomflag
+#define ctm_raise	ct_val.ctrl_image.ct_raise
+#define ctm_rotate	ct_val.ctrl_image.ct_rotate
 #define ctd_colors	ct_val.ctrl_grad.colors
 #define ctd_g_colors	ct_val.ctrl_grad.ct_g_colors
 #define ctd_numcolor	ct_val.ctrl_grad.ct_numcolor
@@ -317,6 +368,10 @@ struct ctrl {
 #define ctic_value	ct_val.ctrl_icon.ct_value
 #define ctic_color	ct_val.ctrl_icon.ct_color
 #define ctic_size	ct_val.ctrl_icon.ct_size
+#define	ctar_xoff	ct_val.ctrl_area.ct_xoff
+#define	ctar_width	ct_val.ctrl_area.ct_width
+#define	ctar_yoff	ct_val.ctrl_area.ct_yoff
+#define	ctar_height	ct_val.ctrl_area.ct_height
 #define ctch_flag	ct_val.ctrl_pcache.ct_cflag
 #define ctch_mode	ct_val.ctrl_pcache.ct_cmode
 #define ctch_effect	ct_val.ctrl_pcache.ct_ceffect
@@ -379,14 +434,23 @@ struct render_state {
 	Drawable target;
 	u_int height;
 	u_int width;
+	int xprefix;
+	int tabxprefix;
+	int xoff;
+	int yoff;
 	u_int ypos;
 	int have_mark;
 	u_int mark_ypos;
 	u_int repaint;
 	int maxascent;
 	int maxdescent;
+	int maxflheight;
+	int max_lineascent;   /* max size above baseline ignoring supscript */
+	int max_linedescent;  /* max size below baseline ignoring subscript */
 	u_int linewidth;
 	u_int brankline;
+	u_int opaque;
+	u_int charoff;
 	struct ctrl *xfont;
 	struct render_object *obj, *objlast;
 };
@@ -402,6 +466,7 @@ struct render_object {
 #define VL_CENTER	1
 #define VL_TOP		2
 #define VL_BOTTOM	3
+#define VL_ICENTER	4
 	enum {
 #ifdef VFLIB
 		O_VFONT,
@@ -414,6 +479,12 @@ struct render_object {
 #ifdef MNG
 		O_ANIM,
 #endif /* MNG */
+#ifdef USE_XFT2
+		O_XTFONT,
+#endif 
+#ifdef USE_M17N
+		O_M17NTEXT,
+#endif 
 		O_ICON
 	} type;
 	u_long fore;
@@ -448,6 +519,23 @@ struct render_object {
 			void *key;
 		} anim;
 #endif /* MNG */
+#ifdef USE_XFT2
+		struct {
+			char *data;
+			char *fontname;
+			char *registry;
+			int len;
+			int size;
+			int charset16;
+		} xftfont;
+#endif
+#ifdef USE_M17N
+		struct {
+			MText *mt;
+			MFrame *drawframe;
+			int from, to;
+		} m17ntext;
+#endif
 	} data;
 };
 
@@ -526,6 +614,7 @@ extern int verbose;
 extern u_int maxpage;
 extern u_int cur_page;
 extern char *mgp_fname;
+extern char *mgp_wname;
 
 extern u_int parse_error;
 extern u_int parse_debug;
@@ -545,6 +634,10 @@ extern struct alloc_color back_clr;
 extern struct alloc_color font_clr;
 
 extern u_int char_size[2];
+extern u_int nonscaled_size[2];
+extern float sup_scale;
+extern float sup_off;
+extern float sub_off;
 extern u_int horiz_gap[2];
 extern u_int vert_gap[2];
 extern u_int depth;
@@ -632,6 +725,7 @@ extern void cleanup __P((int));
 #ifdef TTY_KEYINPUT
 extern void try_enable_ttykey __P((void));
 #endif
+extern void reset_background_pixmap();
 
 /*draw.c*/
 extern void state_goto __P((struct render_state *, u_int, int));
@@ -643,6 +737,14 @@ extern Bool draw_one __P((struct render_state *, XEvent *));
 extern void timebar __P((struct render_state *));
 extern void draw_reinit __P((struct render_state *));
 extern int get_regid __P((char *));
+extern void draw_line_itemsize __P((struct render_state *, int, int, int));
+extern void draw_line_start __P((struct render_state *));
+extern void draw_line_end __P ((struct render_state *));
+
+#ifdef USE_M17N
+extern int obj_new_mtext __P((struct render_state *, int, int,
+			      MText *, int, int, MFrame *, int, int));
+#endif
 
 /*parse.c*/
 extern void load_file __P((char *));
@@ -656,6 +758,7 @@ extern void ctlinsert __P((struct ctrl **, struct ctrl *));
 extern struct ctrl *ctlalloc1 __P((u_int));
 extern void ctlfree __P((struct ctrl *));
 extern struct ctrl *ctlcopy __P((struct ctrl *));
+extern struct ctrl *ctlcopy1 __P((struct ctrl *));
 extern void debug0 __P((struct ctrl *));
 extern void debug1 __P((struct ctrl *));
 
@@ -686,6 +789,7 @@ extern int get_color __P((char *, u_long *));
 extern struct g_color *name2gcolor __P((char *));
 extern void regist_alloc_colors __P((struct alloc_color *, u_long *, int));
 extern void free_alloc_colors __P((struct alloc_color *));
+extern void toggle_fullscreen();
 
 /* background.c */
 extern double cdist __P((int, int, int, int, int, int));
@@ -703,6 +807,9 @@ extern struct tfont *tfc_get __P((u_int, u_int, int, char *, int));
 extern void tfc_setfont __P((char *, int, char *));
 extern XImage *tfc_image __P((struct tfont *, long, long, XImage *, int, int));
 
+/* unimap.c */
+extern void latin_unicode_map_init();
+
 /* embed.c */
 extern char *embed_fname __P((char *));
 extern void embed_file __P((FILE *, struct ctrl *, int *));
@@ -731,3 +838,12 @@ extern void usleep __P((u_int));
 #define COLOR_BUGFIX
 #endif
 
+#ifdef USE_M17N
+/*m17n.c*/
+extern void M17N_init __P((void));
+extern void M17N_set_font __P((char *, char *));
+extern void M17N_set_color __P((u_long));
+extern void M17N_process_direc __P((char *, char *));
+extern void M17N_draw_string __P((struct render_state *, struct ctrl *cp));
+extern void M17N_draw_object __P((struct render_object *, Drawable, int, int));
+#endif
