@@ -26,7 +26,7 @@
  * SUCH DAMAGE.
  */
 /*
- * $Id: parse.c,v 1.86 2001/09/17 14:48:30 nishida Exp $
+ * $Id: parse.c,v 1.102 2007/01/12 10:38:56 nishida Exp $
  */
 
 #include "mgp.h"
@@ -42,6 +42,7 @@ static void thirdpass __P((void));
 static void debug __P((void));
 static int define_font __P((struct ctrl *));
 static struct ctrl *find_font __P((char *));
+static char* clear_esc __P((char*));
 
 /*image*/
 extern int findImage __P((char *, char *));
@@ -133,6 +134,109 @@ cleanup_file()
 	}
 }
 
+/*
+ * clear the given data from escape sequences, so identifying
+ * a character by its hexadecimal value is possible by using \xHH
+ */
+static char*
+clear_esc(s)
+	char* s;
+{
+	char* p;
+	int len;
+	int i;
+	int j;
+	int k;
+	int error = 0;
+	int unicode = 0;
+	int chars;
+	unsigned int code;
+
+	i = j = 0;
+	len = strlen(s);
+	p = (char*)malloc(len + 1);
+
+	while (i < len)
+	{
+		code = 0x0;
+
+		/*
+		 * expecting either beginning of unicode or
+		 * an ascii-character (may be an escape sequence)
+		 */
+		if (!unicode) {
+			/* beginning of unicode */
+			if (s[i] == 0x1b && s[i+1] == '$' &&
+				'@' <= s[i+2] && s[i+2] < 'C') {
+				unicode = 1;
+
+				for (chars = 0; chars < 3; chars++)
+					p[j++] = s[i++];
+
+				continue;
+			}
+
+			/* backslash */
+			if (s[i] == 0x5c) {
+				i++;
+	
+				/* any escape sequence */
+				if (s[i] != 0x5c) {
+
+					 /* hexadecimal token: \xHH */
+					if ((s[i] == 'x' || s[i] == 'X') && i+2 < len) {
+						i++;
+	
+						if (isxdigit(s[i]) && isxdigit(s[i+1])) {
+							for (k = 2; k > 0; k--, i++) {
+								code <<= 4;
+	
+								if (isdigit(s[i]))
+									code += s[i] - 48;
+								else
+									code += tolower(s[i]) - 87;
+							}
+						} else
+							error = 1;
+					} else
+						error = 1;
+				}
+			}
+			if (error)
+				return (char*) NULL;
+
+			if (code == 0x0)
+				p[j++] = s[i++];
+			else
+				p[j++] = code;
+		} else {
+			/*
+			 * expecting either the end of unicode or
+			 * a part of a unicode character
+			 */
+
+			/* end of unicode */
+			if (s[i] == 0x1b && s[i+1] == '(' &&
+				(s[i+2] == 'B' || s[i+2] == 'J')) {
+				unicode = 0;
+
+				for (chars = 0; chars < 3; chars++)
+					p[j++] = s[i++];
+				
+				continue;
+			}
+
+			/* just take the unicode character */
+			p[j++] = s[i++];
+		}
+	}
+	p[j] = 0x0;
+	strcpy(s, p);
+	free(p);
+
+	return s;
+}
+	
 static struct ctrl *
 parse_text(p, page)
 	char *p;
@@ -355,7 +459,7 @@ read_file(fp, filename, page, line, preamble)
 
 		case CTL_TAB:
 		    {
-			int i = root->cti_value - 1;
+			int i = root->cti_value;
 			if (i < 0) {
 				fprintf(stderr, "%s:%d: "
 					"invalid tab index %d\n",
@@ -437,7 +541,8 @@ command:
 			continue;
 		}
 	    {
-		char *p, *q;
+		char *p, *q, *s;
+		int cnt;   /* counter for backslashes at the end of a line */
 
 		p = buf + strlen(buf);
 		if (buf < p && p[-1] == '\n') {
@@ -445,6 +550,22 @@ command:
 			*p = '\0'; 
 		}
 		while (buf < p && p - buf < sizeof(buf) && p[-1] == '\\') {
+
+			/*
+			 * check for quoted backslashes at the end of a line:
+			 * if it is so, then do NOT ignore the line break and
+			 * continue as usual
+			 */
+			s = p;
+			cnt = 0;
+			while (buf < s && s[-1] == '\\') {
+				s--;
+				cnt++;
+			}
+
+			if (cnt % 2 == 0)
+				break;
+
 			p--;
 			if (fgets(buf2, sizeof(buf) - (p - buf), fp) == NULL)
 				break;
@@ -453,6 +574,7 @@ command:
 			while (*q && isspace(*q))
 				q++;
 			strncpy(p, q, sizeof(buf) - (p - buf));
+			p[sizeof(buf) - (p - buf) -1] = '\0';
 			p += strlen(p);
 			if (buf < p && p[-1] == '\n') {
 				p--;
@@ -460,9 +582,9 @@ command:
 			}
 		}
 	    }
-
 		if (buf[0] == '#')
 			continue;
+
 		if (buf[0] == '%') {
 			/* this is directive */
 			int pb;
@@ -479,7 +601,7 @@ command:
 			lex_init(buf + 1);
 			if (yyparse() || n_errors) {
 				fprintf(stderr,
-					"%s:%d: fatal syntax error detected\n",
+					"%s:%d: fatal syntax error detected !!!\n",
 					filename, lineno);
 				exit(-1);
 			}
@@ -661,6 +783,27 @@ command:
 			}
 		} else {
 			/* this is data */
+
+			/*
+			 * escape # and % by backslash (at the beginning
+			 * of a line)
+			 */
+			if (buf[0] == 0x5c && (buf[1] == 0x23 ||
+						buf[1] == 0x25))
+				memcpy(&buf[0], &buf[1], strlen(buf));
+
+			/*
+			 * clear escape sequences
+			 */
+#if 1
+			if (clear_esc(buf) == NULL) {
+				fprintf(stderr, "%s:%d: unknown escape"
+						" sequence\n",
+						filename, lineno);
+				exit(-1);
+			}
+#endif
+
 			cp = parse_text(buf, *page);
 			if (cp) {
 				ch = &page_control[*page][*line];
@@ -741,8 +884,10 @@ secondpass()
 #ifdef MNG
 					cp->ct_op == CTL_ANIM   || 
 #endif /* MNG */
-					cp->ct_op == CTL_PAUSE) 
+					(cp->ct_op == CTL_PAUSE 
+					 	&& l != page_attribute[page].pg_linenum)) {
 						text = 0; 
+					}
 				}
 		}
 		page_attribute[page].pg_text = text;
@@ -773,6 +918,7 @@ thirdpass()
 	struct ctrl *cp;
 	struct ctrl *cp1;
 	struct ctrl *cp2;
+	struct ctrl *cpt;
 	char *p;
 	struct ctrl *carryover[MAXPAGE];
 
@@ -828,6 +974,28 @@ thirdpass()
 	tmpstr[5] = ctlalloc1(CTL_XFONT2);
 	tmpstr[5]->ctc2_value1 = strdup("k14");
 	tmpstr[5]->ctc2_value2 = strdup("jisx0208.1983-*");
+#ifdef USE_M17N
+	if (! (mgp_flag & FL_NOM17N)) {
+		tmpstr[6] = ctlalloc1(CTL_M17N);
+		tmpstr[6]->ctc2_value1 = strdup("fontset");
+		tmpstr[6]->ctc2_value2 = NULL;
+		tmpstr[7] = ctlalloc1(CTL_M17N);
+		tmpstr[7]->ctc2_value1 = strdup("family");
+		tmpstr[7]->ctc2_value2 = NULL;
+		tmpstr[8] = ctlalloc1(CTL_M17N);
+		tmpstr[8]->ctc2_value1 = strdup("language");
+		tmpstr[8]->ctc2_value2 = NULL;
+	} else
+#endif
+	{
+		tmpstr[6] = ctlalloc1(CTL_NOOP);	
+		tmpstr[6]->ctc_value = strdup("");
+		tmpstr[7] = ctlalloc1(CTL_NOOP);	
+		tmpstr[7]->ctc_value = strdup("");
+		tmpstr[8] = ctlalloc1(CTL_NOOP);	
+		tmpstr[8]->ctc_value = strdup("");
+	}
+
 	tmplong[0] = ctlalloc1(CTL_FORE);
 	get_color(DEFAULT_FORE, &tmplong[0]->ctl_value);
 	tmplong[1] = ctlalloc1(CTL_BACK);
@@ -842,6 +1010,10 @@ thirdpass()
 	tmpint[2]->cti_value = DEFAULT_VGAP;
 	tmpint[3] = ctlalloc1(CTL_QUALITY);
 	tmpint[3]->cti_value = DEFAULT_BQUALITY;
+#ifdef USE_XFT2
+	tmpint[4] = ctlalloc1(CTL_OPAQUE);
+	tmpint[4]->cti_value = DEFAULT_OPAQUE;
+#endif
 	tmpvoid[0] = ctlalloc1(CTL_LEFT);
 	tmparea[0] = ctlalloc1(CTL_AREA);
 	tmparea[0]->ctar_width = 100;
@@ -957,13 +1129,14 @@ thirdpass()
 				case CTL_HGAP: tmpint[1] = cp; break;
 				case CTL_VGAP: tmpint[2] = cp; break;
 				case CTL_QUALITY: tmpint[3] = cp; break;
+				case CTL_OPAQUE: tmpint[4] = cp; break;
 
 				case CTL_LEFT: tmpvoid[0] = cp; break;
 				case CTL_RIGHT: tmpvoid[0] = cp; break;
 				case CTL_CENTER: tmpvoid[0] = cp; break;
 				case CTL_LEFTFILL: tmpvoid[0] = cp; break;
 
-				case CTL_AREA: tmpvoid[0] = cp; break;
+				case CTL_AREA: tmparea[0] = cp; break;
 				}
 			}
 		}
@@ -1034,8 +1207,20 @@ thirdpass()
 			continue;
 		line = page_attribute[page].pg_linenum;
 		for (l = 0; l <= line; l++) {
+			int contseen;
+			contseen = 0;	
 			ch = &page_control[page][l];
-			if (default_control[l]) {
+			/* 
+			 * if this line contains CTL_CONT, we don't add 
+			 * default directive to this line 
+			 */
+			for (cp = page_control[page][l]; cp; cp = cp->ct_next) {
+				if (cp->ct_op == CTL_CONT) {
+					contseen++;
+					break;
+				}
+			}
+			if (default_control[l] && !contseen) {
 				ctlinsert(ch, ctlcopy(default_control[l]));
 			}
 		}
@@ -1153,13 +1338,21 @@ thirdpass()
 				}
 			}
 		}
+		if (contseen){
+			/* we have extra cont in the last line */
+			for (cp2 = page_control[page][line-1];
+				cp2->ct_next; cp2 = cp2->ct_next); 
+			cp = ctlalloc1(CTL_LINEEND);
+			if (cp) ctlinsert(&(cp2->ct_next),  cp);
+		}
 	}
     }
 
-	/* insert CTL_TAB items into CTL_LINESTART */
+	/* insert CTL_TAB */
 	for (page = 1; page <= maxpage; page++) {
 		line = page_attribute[page].pg_linenum;
 		for (l = 0; l <= line; l++) {
+			int tab_depth = 0;
 			/*
 			 * if we don't have CTL_LINESTART, we don't add
 			 * directives here.
@@ -1182,10 +1375,8 @@ thirdpass()
 				continue;
 			p = cp->ctc_value;
 			if (p && *p == '\t') {
-				int tab_depth;
-
-				tab_depth = 0;
 				p++;
+				tab_depth++;
 				while (*p == '\t') {
 					tab_depth++;
 					p++;
@@ -1197,12 +1388,30 @@ thirdpass()
 					p = cp->ctc_value = strdup(p);
 					free(tmp);
 				}
+			}
 
-				if (tab_control[tab_depth]) {
-					ctlinsert(&cp1->ct_next,
-					    ctlcopy(tab_control[tab_depth]));
+#if 0
+			/* insert CTL_TAB items into CTL_LINESTART */
+			if (tab_control[tab_depth]) {
+				ctlinsert(&cp1->ct_next,
+				    ctlcopy(tab_control[tab_depth]));
+			}
+#else
+			if (tab_control[tab_depth]) {
+				ch = &page_control[page][l];
+				for (cpt = tab_control[tab_depth]; cpt; cpt = cpt->ct_next){
+					/* Thses ctrl items should be in CTL_LINESTART */
+					if (cpt->ct_op == CTL_IMAGE || cpt->ct_op == CTL_TABPREFIX ||
+							cpt->ct_op == CTL_ICON){
+						ctlinsert(&cp1->ct_next, ctlcopy1(cpt));
+						cp1 = cp1->ct_next; 
+					} else
+					/* other ctrl items should be in the head of the ctrls */
+						ctlinsert(ch, ctlcopy1(cpt));
 				}
 			}
+#endif
+
 			/* special: style escape */
 			if (p && *p == '&') {
 				char *p0;
@@ -1368,6 +1577,7 @@ debug0(p)
 		break;
 	case CTL_SYSTEM:
 	case CTL_XSYSTEM:
+	case CTL_TSYSTEM:
 	case CTL_FILTER:
 		fprintf(stderr, "argc=%d term=%s flag=%d",
 			p->cta_argc,
@@ -1398,6 +1608,17 @@ debug0(p)
 		fprintf(stderr, "type=%s color=%x siz=%d",
 			p->ctic_value, (int)p->ctic_color, (int) p->ctic_size);
 		break;
+	case CTL_VALIGN:
+		fprintf(stderr, ((p->cti_value == VL_CENTER) ? "center"
+				 : ((p->cti_value == VL_TOP) ? "top"
+				    : ((p->cti_value == VL_BOTTOM) ? "bottom"
+				       : "???"))));
+		break;
+	case CTL_AREA:
+		fprintf(stderr, "xoff=%d w=%d yoff=%d h=%d",
+			p->ctar_xoff, p->ctar_width,
+			p->ctar_yoff, p->ctar_height);
+		break;
 	default:
 		fprintf(stderr, "???");
 	}
@@ -1450,7 +1671,7 @@ chkfile(p)
 			fprintf(stderr, "File %s found\n", p);
 		return 0;
 	}
-	fprintf(stderr, "File %s can not found\n", p);
+	fprintf(stderr, "File %s not found\n", p);
 	parse_error++;
 	return -1;
 }
@@ -1556,6 +1777,19 @@ ctlcopy(this)
 	return dst0;
 }
 
+struct ctrl *
+ctlcopy1(src)
+	struct ctrl *src;
+{
+	struct ctrl *dst;
+	if (!src)
+		return NULL;
+	dst = ctlalloc1(0);
+	memcpy(dst, src, sizeof(struct ctrl));
+	dst->ct_next = NULL;
+	return dst;
+}
+
 int
 ctlcmp(a, b)
 	struct ctrl *a;
@@ -1601,6 +1835,7 @@ ctlcmp(a, b)
 		return (a->cti_value == b->cti_value) ? 0 : 1;
 	case CTL_SYSTEM:
 	case CTL_XSYSTEM:
+	case CTL_TSYSTEM:
 	case CTL_FILTER:
 		return 1;
 	case CTL_IMAGE:
