@@ -26,19 +26,26 @@
  * SUCH DAMAGE.
  */
 /*
- * $Id: print.c,v 1.97 2001/09/17 14:48:30 nishida Exp $
+ * $Id: print.c,v 1.109 2007/01/21 09:35:57 nishida Exp $
  */
 /*
  * Paper size selection code is based on psutil.c by Angus J. C. Duggan
  */
 
 #include "mgp.h"
+#ifdef USE_SETLOCALE
+#ifdef HAVE_LOCALE_H
+#include <locale.h>
+#endif
+#endif
 
 static u_int align = AL_LEFT;
 static char *curprefix = NULL;
 static char *tabprefix = NULL;
 static u_int lineheight;
 static u_int linewidth;
+static int char_off = 0;	/* y-offset of the chars (relative to the
+				   current line, used for super/subscript */
 
 #define DEFAULT_PAPER_SIZE	"a4"
 #define PRINT_ASCIIFONT		"Times-Roman"
@@ -52,11 +59,15 @@ static char outputfile[MAXVALLEN];
 static int colorps = 0;
 static int painticon = 0;
 static u_int curlinenum;
-static u_int curpagenum;
+static u_int curpagenum=0;
 static u_int xprefix = 0;
 static u_int tabxprefix = 0;
 static u_int valign = VL_BOTTOM;
 
+/* by Arnd Schmitter 23.07.2004 */
+static int lg; /* Counts the Recursionlevel in print_page() for new Pause-Mode */
+static int PauseMode=0; /* Should Pause-Statements splitet into multiple Pages ? */
+
 #ifdef FREETYPE
 static u_int usetfont = 0;
 #endif
@@ -73,6 +84,7 @@ static int nimagepool = 0;
 static struct textpool {
 	int pfx;
 	int xoffset;
+	int yoffset;     /* offset for super/subscript */
 	int xsiz;
 	struct fontmap *font;
 	int size;
@@ -114,6 +126,10 @@ static struct fontmap {
 	{ CTL_VFONT, KANJI, "*",		"Ryumin-Light-H" },
 #endif /*VFLIB*/
 #ifdef FREETYPE
+	{ CTL_TFONT, ASCII, "schlbk.ttf",	"NewCenturySchlbk-Roman" },
+	{ CTL_TFONT, ASCII, "schlbki.ttf",	"NewCenturySchlbk-Italic" },
+	{ CTL_TFONT, ASCII, "schlbkbd.ttf",	"NewCenturySchlbk-Bold" },
+	{ CTL_TFONT, ASCII, "schlbkbi.ttf",	"NewCenturySchlbk-BoldItalic" },
 	{ CTL_TFONT, ASCII, "times.ttf",	"Times-Roman" },
 	{ CTL_TFONT, ASCII, "timesi.ttf",	"Times-Italic" },
 	{ CTL_TFONT, ASCII, "timesbd.ttf",	"Times-Bold" },
@@ -126,6 +142,7 @@ static struct fontmap {
 	{ CTL_TFONT, ASCII, "couri.ttf",	"Courier-Oblique" },
 	{ CTL_TFONT, ASCII, "courbd.ttf",	"Courier-Bold" },
 	{ CTL_TFONT, ASCII, "courbi.ttf",	"Courier-BoldOblique" },
+	{ CTL_TFONT, ASCII, "symbol.ttf",	"Symbol" },
 	{ CTL_TFONT, ASCII, "*",		"Helvetica" },	/*last resort*/
 #ifdef FREETYPE_CHARSET16
 	{ CTL_TMFONT, KANJI, "wadalab-gothic.ttf", "GothicBBB-Medium-H" },
@@ -215,12 +232,42 @@ static const char *latin1def =
 " /eth /ntilde /ograve /oacute /ocircumflex /otilde /odieresis /divide /oslash /ugrave\n"
 " /uacute /ucircumflex /udieresis /yacute /thorn /ydieresis] /isolatin1encoding exch def\n";
 
+static const char *latin2def = 
+"[ /.notdef /.notdef /.notdef /.notdef /.notdef /.notdef /.notdef /.notdef /.notdef /.notdef\n"
+" /.notdef /.notdef /.notdef /.notdef /.notdef /.notdef /.notdef /.notdef /.notdef /.notdef\n"
+" /.notdef /.notdef /.notdef /.notdef /.notdef /.notdef /.notdef /.notdef /.notdef /.notdef\n"
+" /.notdef /.notdef /space /exclam /quotedbl /numbersign /dollar /percent /ampersand /quoteright\n"
+" /parenleft /parenright /asterisk /plus /comma /minus /period /slash /zero /one\n"
+" /two /three /four /five /six /seven /eight /nine /colon /semicolon\n"
+" /less /equal /greater /question /at /A /B /C /D /E\n"
+" /F /G /H /I /J /K /L /M /N /O\n"
+" /P /Q /R /S /T /U /V /W /X /Y\n"
+" /Z /bracketleft /backslash /bracketright /asciicircum /underscore /quoteleft /a /b /c\n"
+" /d /e /f /g /h /i /j /k /l /m\n"
+" /n /o /p /q /r /s /t /u /v /w\n"
+" /x /y /z /braceleft /bar /braceright /asciitilde /.notdef /.notdef /.notdef\n"
+" /.notdef /.notdef /.notdef /.notdef /.notdef /.notdef /.notdef /.notdef /.notdef /.notdef\n"
+" /.notdef /.notdef /.notdef /.notdef /.notdef /.notdef /.notdef /.notdef /.notdef /.notdef\n"
+" /.notdef /.notdef /.notdef /.notdef /.notdef /.notdef /.notdef /.notdef /.notdef /.notdef\n"
+" /space /Aogonek /breve /Lslash /currency /Lcaron /Sacute /section /dieresis /Scaron\n"
+" /Scedilla /Tcaron /Zacute /hyphen /Zcaron /Zdotaccent /degree /aogonek /ogonek /lslash\n"
+" /acute /lcaron /sacute /caron /cedilla /scaron /scedilla /tcaron /zacute /hungarumlaut\n"
+" /zcaron /zdotaccent /Racute /Aacute /Acircumflex /Abreve /Adieresis /Lacute /Cacute /Ccedilla\n"
+" /Ccaron /Eacute /Eogonek /Edieresis /Ecaron /Iacute /Icircumflex /Dcaron /Dbar /Nacute\n"
+" /Ncaron /Oacute /Ocircumflex /Ohungarumlaut /Odieresis /multiply /Rcaron /Uring /Uacute /Uhungarumlaut\n"
+" /Udieresis /Yacute /Tcedilla /germandbls /racute /aacute /acircumflex /abreve /adieresis /lacute\n"
+" /cacute /ccedilla /ccaron /eacute /eogonek /edieresis /ecaron /iacute /icircumflex /dcaron\n"
+" /dbar /nacute /ncaron /oacute /ocircumflex /ohungarumlaut /odieresis /divide /rcaron /uring\n"
+" /uacute /uhungarumlaut /udieresis /yacute /tcedilla /dotaccent] /isolatin2encoding exch def\n";
+
 static struct encoding {
 	char *name;
 	char *defstr;
 } encodings[] = {
 	{ "iso-8859-1",	"isolatin1encoding", },
 	{ "latin1",	"isolatin1encoding", },
+	{ "iso-8859-2",	"isolatin2encoding", },
+	{ "latin2",	"isolatin2encoding", },
 	{ NULL, NULL, },
 };
 struct encoding *encoding = NULL;
@@ -236,7 +283,8 @@ static void scan_font __P((u_int));
 static void print_page __P((u_int));
 static void print_init __P((void));
 static void print_pageinit __P((u_int));
-static void process_direc_print __P((struct ctrl *));
+static void process_direc_print __P((struct ctrl *cp));
+static void scan_backcolor __P((u_int));
 static void line_start __P((void));
 static void line_end __P((void));
 static void icon_output __P((struct textpool *));
@@ -250,7 +298,7 @@ static void print_fragment __P((u_char *, u_int, int, int));
 static void text_remember __P((char *, int, int, int, int));
 static void icon_remember __P((int, int, int, u_long));
 static void image_remember __P((struct ctrl *, struct imagepool *));
-static void image_load_print __P((char *, int, int, int, int, int));
+static void image_load_print __P((char *, int, int, int, int, int, int));
 static void print_usage __P((char *));
 static void setpencolor __P((u_long));
 static void print_full_image __P((XImage *, Visual *));
@@ -261,6 +309,7 @@ static struct fontmap *findfont __P((int, int, char *));
 static struct fontmap *findttfont __P((int, int, char *));
 static char *fontstring __P((struct fontmap *));
 static void loadfont __P((struct fontmap *, char *));
+static int  count_pause __P((u_int page));   /* Added 23.07.2004 by Arnd Schmitter */
 
 /* image/ *.[ch] */
 extern char *Paths[];
@@ -293,6 +342,9 @@ print_out()
 	window_height = w_height;
 
 	char_size[0] = window_height * DEFAULT_CHARSIZE / 100;
+	sup_off = DEFAULT_SUPOFF;
+	sub_off = DEFAULT_SUBOFF;
+	sup_scale = DEFAULT_SUPSCALE;
 
 	if (outputfile[0]) {
 		if ((fp = fopen(outputfile, "w")) == NULL) {
@@ -307,7 +359,10 @@ print_out()
 		scan_font(i);
 	}
 	for (i = 1; i <= maxpage; i ++)
-		print_page(i);
+	  {
+	    lg=0; /* Set the recursionlevel to 0 for each new page (Arnd Schmitter 23.07.2004) */
+	    print_page(i);
+	  }
 
 	fclose(fp);
 
@@ -380,7 +435,6 @@ scan_font(page)
 			case CTL_XFONT:
 				fprintf(stderr, "obsolete directive XFONT\n");
 				exit(1);
-
 			case CTL_TEXT:
 				if (strstr(cp->ctc_value, "\033$B")
 				 || strstr(cp->ctc_value, "\033$@")) {
@@ -401,7 +455,7 @@ scan_font(page)
 "unable to find proper font for %s\n", "ascii");
 					exit(1);
 				}
-				break;
+
 			default:
 				break;
 			}
@@ -411,6 +465,32 @@ scan_font(page)
 	memset(curfont, 0, sizeof(curfont));
 }
 
+/*
+ * This Function counts the Number of Pause-Statements.
+ */
+static int 
+count_pause(page)
+     u_int page;
+{
+  int count=0;
+  u_int line;
+  struct ctrl*cp;
+  for (line = 0;
+       line <= page_attribute[page].pg_linenum;
+       line++) {
+    curlinenum = line;
+    /* process normal control */
+    for (cp = page_control[page][line]; cp; cp = cp->ct_next)
+      {
+	if(cp->ct_op == CTL_PAUSE)
+	  {
+	    count++;
+	  }
+      }
+  }
+  return count;
+}
+
 static void
 print_page(page)
 	u_int page;
@@ -418,9 +498,21 @@ print_page(page)
 	u_int line;
 	struct ctrl *cp;
 
-	curpagenum = page;
+	int ret=0;  /* A Pause Command was found */
+	int lc = 0; /* Counter to prevent infinite recursion loops at Pause-Statements */
+	int pcount; /* Stores the Number of Pause Statements for a Page */
 
+
+	/* Modified by Arnd Schmitter 23.07.2004:
+     * when in PauseMode for emulationg %pause in PS file (option "-m")
+     * this function is called recursively to print the beginning of
+     * the page again up to current %pause statement
+     */
+	/*curpagenum = page;*/
+	curpagenum++;
+	scan_backcolor(page);
 	print_pageinit(curpagenum);
+	/* End of Modification */
 
 	for (line = 0;
 	     line <= page_attribute[page].pg_linenum;
@@ -428,11 +520,32 @@ print_page(page)
 		curlinenum = line;
 		/* process normal control */
 		for (cp = page_control[page][line]; cp; cp = cp->ct_next)
-			process_direc_print(cp);
+		  {
+		    if ((PauseMode) && (cp->ct_op == CTL_PAUSE))
+		      if (lc >= lg)
+			ret=1;
+		      else 
+			(lc)++;
+		    else 
+		      process_direc_print(cp);
+		    
+		    if (ret) goto pause; /* If a Pause has been found exit Loop immedeatly */
+		  }
 	}
-
+ pause:
 	fprintf(fp, "grestore\n\n");
 	fprintf(fp, "showpage\n\n");
+	/* If a Pause has been found -> Restart this Page */
+	if (ret)
+	  {
+	    lg++;
+	    /* If we are in Pause-Mode check if we are called because of the last CTL_PAUSE at End of Page */
+	    /*   At the Endo of the Last Page is no PAUSE Statemtent. So we have to use all Statements in the
+	     Page */
+	    pcount=count_pause(page);
+	    if ((pcount != lg)||(page == (maxpage)))
+	      print_page(page);
+	  }
 }
 
 static void
@@ -441,10 +554,25 @@ print_init()
 	char **p;
 	char *aligns[] = { "center", "left", "right", NULL };
 
-	fprintf(fp, "%%!PS-Adobe-2.0 EPSF-2.0\n");
+
+	/* Added by Arnd Schmitter 23.07.2004 */
+	/*   If we split Pages with Pause Statements the number of absolute Pages gets greater then maxpage */
+	u_int i;
+	int absolute_pages = maxpage+1;
+	if (PauseMode)
+	  {
+	    for (i = 0; i < maxpage; i++)
+	      {
+		absolute_pages+=count_pause(i)-1;
+	      }	   
+	  }
+
+	/*fprintf(stderr,"Pages %d\n",absolute_pages);*/
+
+	fprintf(fp, "%%!PS-Adobe-2.0\n");
 	fprintf(fp, "%%%%Creator: mgp2ps\n");
 	fprintf(fp, "%%%%Title: %s\n", mgp_fname);
-	fprintf(fp, "%%%%Pages: %d\n", maxpage);
+	fprintf(fp, "%%%%Pages: %d\n", absolute_pages);
 	fprintf(fp, "%%%%BoundingBox: 0 0 %d %d\n", paper->height, paper->width);
 	fprintf(fp, "%%%%DocumentPaperSizes: %s\n", paper->name);
 	fprintf(fp, "%%%%Orientation: Landscape\n");
@@ -460,7 +588,12 @@ print_init()
 	fprintf(fp, "/resety {/ymin 0 def /ymax 0 def} def\n");
 	fprintf(fp, "/setymax {dup ymax gt {dup /ymax exch def} if pop} def\n");
 	fprintf(fp, "/setymin {dup ymin lt {dup /ymin exch def} if pop} def\n");
-	fprintf(fp, "/calcy {false charpath flattenpath pathbbox setymax pop setymin pop} def\n");
+	fprintf(fp, "/calcy {dup ( ) eq\n"
+		"  {0 setymax 0 setymin pop}\n"
+		"  {pop charsize 0.7 mul setymax charsize -0.15 mul setymin}\n"
+		"  ifelse\n"
+		"} def\n");
+
 	fprintf(fp, "resety\n");
 
 	/* define writebox */
@@ -479,7 +612,7 @@ print_init()
 	fprintf(fp, "  WIDTH -1 mul 0 rlineto closepath eofill stroke\n");
 	fprintf(fp, "} def\n");
 
-	/* ypos = ypos - (charsize * (1 + vert_gap / 100)) */
+	/* ypos = ypos - (charsize * (1 + vert_gap / 100)) + little margin */
 	fprintf(fp, "/NL {\n");
 	fprintf(fp, "  charsize imgsize gt\n");
 	fprintf(fp, "	ymax ymin sub /csize exch def\n");
@@ -488,6 +621,7 @@ print_init()
 	fprintf(fp, "    { vertgap 100 div csize mul imgsize add }\n");
 	fprintf(fp, "  ifelse\n");
 	fprintf(fp, "  ypos exch sub /ypos exch def\n");
+	fprintf(fp, "  HEIGHT 90 div ypos exch sub /ypos exch def\n");
 	fprintf(fp, "} bind def\n");
 
 	/* show with line wrapping */
@@ -579,6 +713,8 @@ print_init()
 	/* emit character encoding definition, if necessary */
 	if (encoding && strcmp(encoding->defstr, "isolatin1encoding") == 0)
 		fprintf(fp, "%s", latin1def);
+	if (encoding && strcmp(encoding->defstr, "isolatin2encoding") == 0)
+		fprintf(fp, "%s", latin2def);
 }
 
 static void
@@ -619,6 +755,7 @@ process_direc_print(cp)
 		fprintf(fp, "%d setcharsize\n", char_size[0]);
 		break;
 
+	  
 #ifdef VFLIB
 	case CTL_VFONT:
 		font = findfont(cp->ct_op, KANJI, cp->ctc_value);
@@ -748,6 +885,13 @@ process_direc_print(cp)
 		print_line(cp->ctc_value, 0);
 		if (align == AL_LEFTFILL0)
 			align = AL_LEFTFILL1;
+
+		/* disable super/subscript for next line */
+		if (char_off != 0)
+		{
+			char_size[0]=nonscaled_size[0];
+			char_off=0;
+		}
 		break;
 	    }
 
@@ -769,9 +913,11 @@ process_direc_print(cp)
 		fore = cp->ctl_value;
 		break;
 
+#if 0 /* it's not necessary */
 	case CTL_BACK:
 		back = cp->ctl_value;
 		break;
+#endif
 
 	/* 
      * MARK and AGAIN processing by A.Ito, 14 Jul. 2000 
@@ -801,12 +947,75 @@ process_direc_print(cp)
 		    (double)cp->ctar_yoff / cp->ctar_height);
 		fprintf(fp, "/xpos 0 def\n");
 		break;
+    case CTL_OPAQUE:
+		if (cp->cti_value > 100){
+			fprintf(stderr, "%%opaque: value should be 0-100\n");
+			cp->cti_value = 100;
+		}
+		fprintf(fp, "%f setgray\n", 1.0 - (float)cp->cti_value /100.0);
+		break;
+
+    /* setup for super/subscript */
+    case CTL_SETSUP:
+                if (cp->cti3_value1 > 100 || cp->cti3_value1 < 10){
+                     sup_off = DEFAULT_SUPOFF;
+                }else{
+                     sup_off = cp->cti3_value1 / 100.;
+                }
+                if (cp->cti3_value2 > 100 || cp->cti3_value2 < 10){
+                     sub_off = DEFAULT_SUBOFF;
+                }else{
+                     sub_off = cp->cti3_value2 / 100.;
+                }
+                if (cp->cti3_value3 > 100 || cp->cti3_value3 < 10){
+                     sup_scale = DEFAULT_SUPSCALE;
+                }else{
+                     sup_scale = cp->cti3_value3 / 100.;
+                }
+                break;
+
+    /* subscript */
+    case CTL_SUB:
+		/* save old size and set new one */
+		nonscaled_size[0] = char_size[0];
+		char_size[0] = char_size[0] * sup_scale;
+		char_off = nonscaled_size[0] * (-sub_off);
+		fprintf(fp, "%d setcharsize\n", char_size[0]);
+		break;
+
+    /* superscript */
+    case CTL_SUP:
+		/* save old size and set new one */
+		nonscaled_size[0] = char_size[0];
+		char_size[0] = char_size[0] * sup_scale;
+		char_off = nonscaled_size[0] * sup_off;
+		fprintf(fp, "%d setcharsize\n", char_size[0]);
+		break;
 
 	default:
-		break;
+	  break;
 	}
 }
 
+static void
+scan_backcolor(u_int page)
+{
+  struct ctrl *cp;
+  u_int line;
+
+	for (line = 0; line <= page_attribute[page].pg_linenum; line++) {
+		for (cp = page_control[page][line]; cp; cp = cp->ct_next)
+			switch(cp->ct_op) {
+				case CTL_BACK:
+					back = cp->ctl_value;
+				break;
+				default:
+				break;
+      		}
+	}
+}
+
+
 static void
 line_start()
 {
@@ -957,7 +1166,17 @@ line_end()
 		if (textpool[i].text) {
 			if (!textpool[i].pfx)
 				fprintf(fp, "updatefillzero ");
+
+			/* for super/subscript: modify the current y-position */
+			if (textpool[i].yoffset != 0)
+				fprintf(fp, "currentpoint %d add moveto ",
+					textpool[i].yoffset);
 			fprintf(fp, "show\n");
+
+			/* for super/subscript: reset the old y-position */
+			if (textpool[i].yoffset != 0)
+				fprintf(fp, "currentpoint %d sub moveto ",
+					textpool[i].yoffset);
 		} else {
 			fprintf(fp, "\n");
 			if (textpool[i].font) {
@@ -994,7 +1213,7 @@ line_end()
 		if (cp1) {
 			image_load_print(cp1->ctm_fname, cp1->ctm_numcolor,
 				cp1->ctm_ximagesize, cp1->ctm_yimagesize, 0,
-				cp1->ctm_zoomflag);
+				cp1->ctm_zoomflag, cp1->ctm_rotate);
 		}
         fprintf(fp, "/xpos xpos%d def xpos ypos moveto\n",
 	        imagepool[i].target_text);
@@ -1019,8 +1238,8 @@ icon_output(tp)
 	int ixoff, iyoff;
 	int paintit;
 
-	iyoff = (csize - isize) / 3;	/*XXX*/
-	ixoff = tp->xoffset + iyoff;
+	iyoff = (csize - isize) / 6;	/*XXX*/
+	ixoff = tp->xoffset + (csize - isize) / 2;  /* XXX */
 
 	paintit = (painticon || colorps);
 
@@ -1028,7 +1247,7 @@ icon_output(tp)
 	case 0:
 		/* XXX: image is not supported yet */
 		break;
-	case 1:
+	case 1: /* this is box */
 		fprintf(fp, "currentpoint ");
 		if (paintit)
 			fprintf(fp, "currentpoint newpath moveto ");
@@ -1042,7 +1261,7 @@ icon_output(tp)
 		fprintf(fp, "stroke moveto\n");
 		fprintf(fp, "%d 0 rmoveto\n", ixoff * 2 + isize);
 		break;
-	case 2:
+	case 2: /* this is arc */
 		fprintf(fp, "currentpoint ");
 		if (paintit)
 			fprintf(fp, "currentpoint newpath moveto ");
@@ -1313,7 +1532,11 @@ print_fragment(data, len, kanjimode, pfx)
 	/* by A.Ito 14 Jul. 2000 */
 	/* placeholder */
 	imagepool[nimagepool].image = NULL;
+#if 0
 	imagepool[nimagepool].target_text = ntextpool-1;
+#else
+	imagepool[nimagepool].target_text = nimagepool;
+#endif
 	nimagepool++;
 }
 
@@ -1360,6 +1583,7 @@ text_remember(text, ctype, fontsize, offset, pfx)
 	}
 	textpool[ntextpool].pfx = pfx;
 	textpool[ntextpool].xoffset = offset;
+	textpool[ntextpool].yoffset = char_off;
 	textpool[ntextpool].xsiz = linewidth - offset;
 	textpool[ntextpool].size = fontsize;
 	textpool[ntextpool].font = curfont[ctype];
@@ -1387,6 +1611,105 @@ icon_remember(icon, fontsize, offset, color)
 	ntextpool++;
 }
 
+
+/* !TODO: move rotation code into some library */
+/* rotate image by 90 degrees (counter clockwise) */
+static void rotate_image_p90(image)
+	Image *image;
+{
+	unsigned int row, column, pl = image->pixlen;
+	unsigned int new_height = image->width, new_width = image->height, new_linelen = new_width * pl;
+	byte *src, *tgt, *col_head;
+	Pixel d;
+	/* allocate buffer for new image */
+	byte *rot_data = lmalloc(new_linelen * new_height);
+
+	/* do the rotation */
+	for (row = 0, src = image->data, col_head = rot_data + (new_height - 1) * new_linelen; 
+			row < image->height; 
+			row++, col_head += pl) {
+		for (column = 0, tgt = col_head; 
+				column < image->width; 
+				column++, src += pl, tgt -= new_linelen) {
+			d = memToVal(src, pl);
+			valToMem(d, tgt, pl);
+		}
+	}
+
+	/* swap to rotated image, exchange height and width
+	   and point to rotated data */
+	image->height = new_height;
+	image->width = new_width;
+	lfree(image->data);
+	image->data = rot_data;
+}
+
+
+/* rotate image by -90 degrees (clockwise) */
+static void rotate_image_m90(image)
+Image *image;
+{
+	unsigned int row, column, pl = image->pixlen;
+	unsigned int new_height = image->width, new_width = image->height, new_linelen = new_width * pl;
+	byte *src, *tgt;
+	Pixel d;
+	/* allocate buffer for new image */
+	byte *rot_data = lmalloc(new_linelen * new_height);
+
+	/* do the rotation */
+	for (row = 0, src = image->data; row < image->height; row++) {
+		for (column = 0, tgt = rot_data + new_linelen - (row + 1) * pl; 
+				column < image->width; 
+				column++, src += pl, tgt += new_linelen) {
+			d = memToVal(src, pl);
+			valToMem(d, tgt, pl);
+		}
+	}
+
+	/* swap to rotated image, exchange height and width
+	   and point to rotated data */
+	image->height = new_height;
+	image->width = new_width;
+	lfree(image->data);
+	image->data = rot_data;
+
+	return;
+}
+
+
+/* rotate image by 180 degrees */
+static void rotate_image_180(image)
+	Image *image;
+{
+	unsigned int row, column, pl = image->pixlen;
+	unsigned int new_height = image->height, new_width = image->width, new_linelen = new_width * pl;
+	byte *src, *tgt;
+	Pixel d;
+	/* allocate buffer for new image */
+	byte *rot_data = lmalloc(new_linelen * new_height);
+
+	/* do the rotation */
+	for (row = 0, src = image->data; row < image->height; row++) {
+		for (column = 0, tgt = rot_data + (new_height - row) * new_linelen - pl; 
+				column < image->width; 
+				column++, src += pl, tgt -= pl) {
+			d = memToVal(src, pl);
+			valToMem(d, tgt, pl);
+		}
+	}
+
+	/* swap to rotated image, exchange height and width
+	   and point to rotated data */
+	image->height = new_height;
+	image->width = new_width;
+	lfree(image->data);
+	image->data = rot_data;
+
+	return;
+}
+
+
+
 static void
 image_remember(cp, pool)
 	struct ctrl *cp;
@@ -1409,6 +1732,29 @@ image_remember(cp, pool)
 			fprintf(stderr, "failed to open %s\n", cp->ctm_fname);
 			exit(1);
 		}
+		switch (cp->ctm_rotate) {
+			case 0:
+				/* Do nothing */
+				break;
+		
+			case -90:
+			case 270:
+				rotate_image_m90(myimage);
+				break;
+
+			case 90:
+				rotate_image_p90(myimage);
+				break;
+
+			case -180:
+			case 180:
+				rotate_image_180(myimage);
+				break;
+
+			default:
+				fprintf(stderr, "rotation by %d degrees not supported.\n", rotate);
+				cleanup(-1);
+		}
 		width = myimage->width;
 		height = myimage->height;
 		freeImage(myimage);
@@ -1435,6 +1781,29 @@ noneps:
 		fprintf(stderr, "failed to open %s\n", cp->ctm_fname);
 		exit(1);
 	}
+	switch (cp->ctm_rotate) {
+		case 0:
+			/* Do nothing */
+			break;
+	
+		case -90:
+		case 270:
+			rotate_image_m90(myimage);
+			break;
+
+		case 90:
+			rotate_image_p90(myimage);
+			break;
+
+		case -180:
+		case 180:
+			rotate_image_180(myimage);
+			break;
+
+		default:
+			fprintf(stderr, "rotation by %d degrees not supported.\n", rotate);
+			cleanup(-1);
+	}
 	pool->xsiz = myimage->width;
 	pool->ysiz = myimage->height;
 	freeImage(myimage);
@@ -1443,13 +1812,14 @@ noneps:
 }
 
 static void
-image_load_print(filename, numcolor, ximagesize, yimagesize, backflag, zoomflag)
+image_load_print(filename, numcolor, ximagesize, yimagesize, backflag, zoomflag, rotate)
 	char *filename;
 	int numcolor;
 	int ximagesize;
 	int yimagesize;
 	int backflag;
 	int zoomflag;
+	int rotate;
 {
 	Image *myimage, *image;
 	Pixmap mypixmap;
@@ -1488,6 +1858,29 @@ image_load_print(filename, numcolor, ximagesize, yimagesize, backflag, zoomflag)
 
 	if ((myimage = loadImage(filename, verbose)) == NULL)
 		exit(-1);	/* fail to load image data */
+	switch (rotate) {
+		case 0:
+			/* Do nothing */
+			break;
+	
+		case -90:
+		case 270:
+			rotate_image_m90(myimage);
+			break;
+
+		case 90:
+			rotate_image_p90(myimage);
+			break;
+
+		case -180:
+		case 180:
+			rotate_image_180(myimage);
+			break;
+
+		default:
+			fprintf(stderr, "rotation by %d degrees not supported.\n", rotate);
+			cleanup(-1);
+	}
 	width = myimage->width;
 	height = myimage->height;
 
@@ -1598,7 +1991,7 @@ static void
 print_usage(name)
 	char	*name;
 {
-	fprintf(stderr, "Usage: %s [-cirVt] [-f psfile] [-p paper] "
+	fprintf(stderr, "Usage: %s [-mcirVt] [-f psfile] [-p paper] "
 		"[-x xmargin] [-y ymargin] [-e encoding] mgpfile\n", name);
 	exit(0);
 }
@@ -1718,9 +2111,9 @@ print_32_image(print_image, visual)
 		for (x = 0; x < width; x ++) {
 			unsigned long pix;
 			pix = XGetPixel(print_image, x, y);
-			r =  pix & visual->red_mask & 0xff;
-			g = (pix & visual->green_mask >> 8)  & 0xff;
-			b = (pix & visual->blue_mask  >> 16) & 0xff;
+			r = ((pix & visual->red_mask)   >> 16) & 0xff;
+			g = ((pix & visual->green_mask) >>  8) & 0xff;
+			b =   pix & visual->blue_mask          & 0xff;
 
 			if (reverse)
 				fprintf(fp, "%02x%02x%02x",
@@ -1907,13 +2300,20 @@ main(argc, argv)
 	char *progname;
 	static char pathbuf[MAXPATHLEN];
 
+#ifdef USE_SETLOCALE
+        setlocale(LC_CTYPE, "");
+#endif
+
 	progname = argv[0];
 
 	/* set default paper size */
 	paper = findpaper(DEFAULT_PAPER_SIZE);
 
-	while ((opt = getopt(argc, argv, "ciVrf:x:y:p:e:Pt")) != -1) {
+	while ((opt = getopt(argc, argv, "mciVrf:x:y:p:e:X:Pt")) != -1) {
 		switch (opt) {
+		case 'm':         /* Added by Arnd Schmitter 23.07.2004 */
+		  PauseMode=1;
+		  break;
 		case 'c':
 			colorps++;
 			break;
@@ -1978,6 +2378,15 @@ main(argc, argv)
 #endif
 			break;
 
+       case 'X':
+			if (strcmp(optarg, "freetype") == 0)
+				mgp_flag |= FL_NOFREETYPE;
+			else {
+				fprintf(stderr, "unknown rendering engine %s\n", optarg);
+			}
+			break;
+
+
 		default:
 			print_usage(progname);
 			/*NOTREACHED*/
@@ -2031,6 +2440,8 @@ findfont(ctrl, lang, font)
 		return NULL;
 
 #ifdef FREETYPE
+	if (mgp_flag & FL_NOFREETYPE && ctrl == CTL_TFONT)
+		return NULL;
 	if (lang == ASCII && usetfont)
 		if ((p = findttfont(ctrl, lang, font))) goto found;
 #endif
@@ -2113,7 +2524,21 @@ loadfont(font, name)
 			font->psfont, font->font);
 	}
 	if (font->lang == ASCII && encoding) {
-		fprintf(fp,
+		/*
+		 * Symbol font does not need latin1-encoding
+		 */
+		if (strcmp(font->psfont, "Symbol") == 0) {
+			fprintf(fp,
+				"/%s%s\n"
+				"	/%s findfont\n"
+				"		dup length dict begin\n"
+				"		{1 index /FID ne {def} {pop "
+				"pop} ifelse} forall\n"
+				"	currentdict end\n"
+				"definefont pop\n",
+				fontstring(font), "t", font->psfont);
+		} else {
+			fprintf(fp,
 "/%s%s\n"
 "	/%s findfont\n"
 "		dup length dict begin\n"
@@ -2122,6 +2547,8 @@ loadfont(font, name)
 "	currentdict end\n"
 "definefont pop\n",
 			fontstring(font), "t", font->psfont, encoding->defstr);
+		}
+
 		fprintf(fp, "/%s {/%s%s findfont exch scalefont setfont} def\n",
 			fontstring(font), fontstring(font), "t");
 	} else {
@@ -2203,8 +2630,7 @@ findttfont(ctrl, lang, font)
 	pfont->ctrl = ctrl;
 	pfont->lang = lang;
 	pfont->font = font;
-	pfont->psfont = (char *)malloc(string_len +1);
-	strncpy(pfont->psfont, string, string_len);
+	pfont->psfont = strdup(string);
 	pfont->fontid = 0;
 	pfont->loaded = 0;
 
@@ -2219,3 +2645,4 @@ findttfont(ctrl, lang, font)
 	return pfont;
 }
 #endif /* FREETYPE */
+
